# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import subprocess

from .io import info, err


def _format_subprocess_command(command):
    return ' '.join(command) if isinstance(command, list) else command


def try_subprocess_call_and_return(command, environ=None):
    """Run `command` in a child process, capturing all stdout before returning."""
    try:
        env = os.environ.copy()
        env.update(environ or {})

        command = _format_subprocess_command(command)
        p = subprocess.Popen(
            command, stdout=subprocess.PIPE, env=env, shell=True
        )
        out, error = p.communicate()

        if p.returncode != 0:
            return False, error
        return True, out.strip('\n')
    except (OSError, subprocess.CalledProcessError) as e:
        return False, e


def try_subprocess_call(command, environ=None):
    """Run `command` in a child process."""
    info('executing cmd: %s' % command)
    try:
        env = os.environ.copy()
        env.update(environ or {})

        command = _format_subprocess_command(command)
        subprocess.check_call([command], env=env, shell=True)
        return True
    except (OSError, subprocess.CalledProcessError) as e:
        err('(subprocess) %s' % e)
        return False


def print_environ(environ):
    info('\n'.join(sorted(['%s=%s' % (k, v) for k, v in environ.iteritems()])))



# -*- coding: utf-8 -*-
import os
import tarfile
import hashlib

from ..config import config as app_conf


def md5sum_dir(dirname, tar_file_path=app_conf.lock.tar_file_path):
    """MD5 hash a directory by first creating a `tar` and hashing the tar."""
    if not os.path.isdir(dirname):
        raise TypeError('"%s" is not a directory' % dirname)

    with tarfile.open(tar_file_path, 'w') as tar:
        tar.add(dirname, arcname=os.path.basename(dirname))

    hash_md5 = hashlib.md5()
    with open(tar_file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b''):
            hash_md5.update(chunk)
    try:
        os.remove(tar_file_path)
    except OSError:
        pass
    return hash_md5.hexdigest()


def read_lock(lock_file_path=app_conf.lock.lock_file_path):
    try:
        with open(lock_file_path, 'rU') as f:
            return f.readline().strip('\n')
    except IOError:
        return None


def write_lock(dirname, lock_file_path=app_conf.lock.lock_file_path):
    with open(lock_file_path, 'w') as f:
        f.write(md5sum_dir(dirname))



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals

import sys

# @see: http://pueblo.sourceforge.net/doc/manual/ansi_color_codes.html
U_ERROR = '\033[1;31m✖ \033[0m'
U_WARNING = '\033[33m⚠ \033[0m'
U_INFO = '\033[34m➛ \033[0m'
U_OK = '\033[32m✓ \033[0m'


def err(message, exit_=True):
    print(U_ERROR + message, file=sys.stderr)
    if exit_:
        exit(1)


def warn(message):
    print(U_WARNING + message, file=sys.stderr)


def info(message, end='\n'):
    message = "{prefix} {message}".format(**{
        'prefix': U_INFO, 'message': message,
    })
    print(message, file=sys.stdout, end=end)


def ok(message):
    print(U_OK + message, file=sys.stdout)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from ..core.utils import try_subprocess_call
from ..config import config as app_conf

import docker

_client = docker.from_env()


def run(arguments, environ):
    command = ['docker', 'run'] + arguments
    command = ' '.join(command)
    return try_subprocess_call(command, environ=environ)


def stop(container_name):
    _client.stop(container_name)
    return True


def get_image(name, tag=app_conf.docker.default_local_tag):
    image_name = '{name}:{tag}'.format(name=name, tag=tag)
    image_list = _client.images()
    found_image = [i for i in image_list if i['RepoTags'] == [image_name]]

    image = next(iter(found_image), None)
    return image_name if image else None


def cleanse_image_tag(tag):
    tag = ''.join([c for c in tag if c not in app_conf.docker.invalid_tag_chars])
    tag = tag.replace('/', '.')
    return tag



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
bling.py

usage:
    bling start
    bling stop
    bling run
    bling build [--publish] [--debug]

commands:
    start         starts app inside an interactive container
    stop          stops running container
    run           starts app inside a container, executing the specified run command
    build         invokes `packer build` with packer template json file

options:
    --publish     publishes the build artifact to a remote registry
    --debug       enables debug mode

    -h --help     shows this
    -v --version  shows version
"""
from . import __version__

from .core import blingrc
from .core import commands
from .core.io import err

from .domain.blingrc import MissingBlingConfig
from .domain.blingrc import MalformedBlingConfig
from .domain.blingrc import InvalidBlingConfig

from docopt import docopt


def _parse_user_input(config, args):
    is_valid = blingrc.validate_config(config)
    if not is_valid:
        err('invalid bling config (see above), exiting')
        return None
    if args['start']:
        return commands.start(config)
    if args['stop']:
        return commands.stop(config)
    if args['build']:
        return commands.build(config, should_publish=args['--publish'], is_debug=args['--debug'])
    if args['run']:
        return commands.run(config)
    err('unexpected bling command..., exiting')


def main():
    args = docopt(__doc__, version=__version__)
    try:
        config = blingrc.open_config()
    except (MissingBlingConfig, MalformedBlingConfig, InvalidBlingConfig) as e:
        return err('%s' % e)
    try:
        _parse_user_input(config, args)
    except KeyboardInterrupt:
        pass
    exit(0)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import os
import json
import shutil

from ..core.lock import md5sum_dir
from ..core.lock import read_lock
from ..domain.vendor import PackerConfigNotFound
from ..config import config as app_conf


def is_packer_updated(packer_dir):
    old_hash = read_lock()
    new_hash = md5sum_dir(packer_dir)
    return old_hash and old_hash != new_hash


def read_config(path, packer_config_file=app_conf.packer.config_file):
    path = os.path.join(path, packer_config_file)
    path = os.path.abspath(path)
    try:
        with open(path, 'rU') as f:
            packer_template = json.load(f)
    except IOError:
        raise PackerConfigNotFound('cannot open "%s"' % path)
    return packer_template


def write_config(config, destination):
    dirname = os.path.dirname(destination)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    with open(destination, 'w') as f:
        json.dump(config, f, indent=2, sort_keys=True)
    return destination


def replace_post_processor(packer_path, name, tag):
    packer_template = read_config(packer_path)
    packer_template = packer_template.copy()
    packer_template['post-processors'] = []
    packer_template['post-processors'].append([
        {
            'repository': name,
            'tag': tag,
            'type': 'docker-import',
        }
    ])
    return packer_template


def bootstrap_config(
        config, should_publish, is_debug,
        default_local_tag=app_conf.docker.default_local_tag,
        packer_tmp_path=app_conf.packer.tmp_path,
        packer_config_file=app_conf.packer.config_file):
    """Update the Packer config file depending on how we want to produce an image."""

    command = ['packer', 'build']
    if is_debug:
        command.append('-debug')
    if should_publish:
        command.append(os.path.join(config.get('packer'), packer_config_file))
    else:
        template = replace_post_processor(
            config.get('packer'),
            config.get('image_name'),
            default_local_tag
        )
        packer_template_path = write_config(template, packer_tmp_path)
        command.append(packer_template_path)
    return command


def cleanup_config(packer_tmp_root=app_conf.packer.tmp_root, packer_tmp_path=app_conf.packer.tmp_path):
    """Removes the temporary packer dumping ground directory if exists."""
    if os.path.exists(packer_tmp_path):
        shutil.rmtree(packer_tmp_root)



# -*- coding: utf-8 -*-


class ConfigValidation(object):
    def __init__(self, config, errors=None):
        self.config = config
        self.errors = errors

    @property
    def has_error(self):
        return bool(self.errors)


class MissingBlingConfig(Exception):
    pass


class MalformedBlingConfig(Exception):
    pass


class InvalidBlingConfig(Exception):
    pass



# -*- coding: utf-8 -*-
import os

import blingrc

from .utils import try_subprocess_call
from .io import err, warn
from .lock import write_lock

from ..vendor import travis
from ..vendor import packer
from ..vendor import docker

from ..config import config as app_conf


def start(config, run_command=app_conf.commands.run_command):
    packer_path = config.get('packer')
    if packer.is_packer_updated(packer_path):
        warn('packer config has changed, re-run bling build')
        return False

    image_name = docker.get_image(config.get('image_name'))
    if not image_name:
        err('docker image for "%s" does not exist, run bling build' % image_name)
        return False

    write_lock(packer_path)
    command = config.get('start') + ['--name', config.get('name'), image_name, run_command]
    environ = blingrc.patch_environ(os.environ, config.get('env'))
    return docker.run(command, environ)


def stop(config):
    return docker.stop(config.get('name'))


def run(config):
    environ = blingrc.patch_environ(os.environ, config.get('env'))
    return try_subprocess_call(config.get('run'), environ=environ)


def build(config, should_publish=False, is_debug=False):
    command = packer.bootstrap_config(config, should_publish, is_debug)

    environ = travis.patch_environ(os.environ)
    environ = travis.cleanse_environ(environ)
    environ = blingrc.patch_environ(environ, config.get('env'))

    is_success = try_subprocess_call(command, environ=environ)

    packer.cleanup_config()
    write_lock(config.get('packer'))
    return is_success



# -*- coding: utf-8 -*-
from ..core.utils import try_subprocess_call_and_return
from ..domain.vendor import GitError


def _execute_command(command):
    is_success, response = try_subprocess_call_and_return(command)
    if not is_success:
        raise GitError('"%s" failed with error code: %s' % (' '.join(command), response))
    return response


def get_current_branch():
    command = ['git', 'rev-parse', '--abbrev-ref', 'HEAD']
    return _execute_command(command)


def get_current_commit():
    command = ['git', 'rev-parse', 'HEAD']
    return _execute_command(command)
